"use strict";

//
// Copyright (c) ZeroC, Inc. All rights reserved.
//
var Ice = require("../Ice/ModuleRegistry").Ice;

require("../Ice/Identity");

require("../Ice/LocalException");

require("../Ice/StringUtil");

var Identity = Ice.Identity;
var IdentityParseException = Ice.IdentityParseException;
var StringUtil = Ice.StringUtil;
/**
* Converts a string to an object identity.
*
* @param s The string to convert.
*
* @return The converted object identity.
**/

Ice.stringToIdentity = function (s) {
  var ident = new Identity(); //
  // Find unescaped separator; note that the string may contain an escaped
  // backslash before the separator.
  //

  var slash = -1;
  var pos = 0;

  while ((pos = s.indexOf('/', pos)) !== -1) {
    var escapes = 0;

    while (pos - escapes > 0 && s.charAt(pos - escapes - 1) == '\\') {
      escapes++;
    } //
    // We ignore escaped escapes
    //


    if (escapes % 2 === 0) {
      if (slash == -1) {
        slash = pos;
      } else {
        //
        // Extra unescaped slash found.
        //
        throw new IdentityParseException("unescaped backslash in identity `".concat(s, "'"));
      }
    }

    pos++;
  }

  if (slash == -1) {
    ident.category = "";

    try {
      ident.name = StringUtil.unescapeString(s, 0, s.length, "/");
    } catch (e) {
      throw new IdentityParseException("invalid identity name `".concat(s, "': ").concat(e.toString()));
    }
  } else {
    try {
      ident.category = StringUtil.unescapeString(s, 0, slash, "/");
    } catch (e) {
      throw new IdentityParseException("invalid category in identity `".concat(s, "': ").concat(e.toString()));
    }

    if (slash + 1 < s.length) {
      try {
        ident.name = StringUtil.unescapeString(s, slash + 1, s.length, "/");
      } catch (e) {
        throw new IdentityParseException("invalid name in identity `".concat(s, "': ").concat(e.toString()));
      }
    } else {
      ident.name = "";
    }
  }

  return ident;
};
/**
* Converts an object identity to a string.
*
* @param ident The object identity to convert.
*
* @param toStringMode Specifies if and how non-printable ASCII characters are escaped in the result.
*
* @return The string representation of the object identity.
**/


Ice.identityToString = function (ident) {
  var toStringMode = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : Ice.ToStringMode.Unicode;

  if (ident.category === null || ident.category.length === 0) {
    return StringUtil.escapeString(ident.name, "/", toStringMode);
  } else {
    return StringUtil.escapeString(ident.category, "/", toStringMode) + '/' + StringUtil.escapeString(ident.name, "/", toStringMode);
  }
};
/**
* Compares the object identities of two proxies.
*
* @param lhs A proxy.
* @param rhs A proxy.
* @return -1 if the identity in <code>lhs</code> compares
* less than the identity in <code>rhs</code>; 0 if the identities
* compare equal; 1, otherwise.
*
* @see ProxyIdentityKey
* @see ProxyIdentityAndFacetKey
* @see ProxyIdentityAndFacetCompare
**/


Ice.proxyIdentityCompare = function (lhs, rhs) {
  if (lhs === rhs) {
    return 0;
  } else if (lhs === null && rhs !== null) {
    return -1;
  } else if (lhs !== null && rhs === null) {
    return 1;
  } else {
    var lhsIdentity = lhs.ice_getIdentity();
    var rhsIdentity = rhs.ice_getIdentity();
    var n = lhsIdentity.name.localeCompare(rhsIdentity.name);
    return n !== 0 ? n : lhsIdentity.category.localeCompare(rhsIdentity.category);
  }
};
/**
* Compares the object identities and facets of two proxies.
*
* @param lhs A proxy.
* @param rhs A proxy.
* @return -1 if the identity and facet in <code>lhs</code> compare
* less than the identity and facet in <code>rhs</code>; 0 if the identities
* and facets compare equal; 1, otherwise.
*
* @see ProxyIdentityAndFacetKey
* @see ProxyIdentityKey
* @see ProxyIdentityCompare
**/


Ice.proxyIdentityAndFacetCompare = function (lhs, rhs) {
  if (lhs === rhs) {
    return 0;
  } else if (lhs === null && rhs !== null) {
    return -1;
  } else if (lhs !== null && rhs === null) {
    return 1;
  } else {
    var lhsIdentity = lhs.ice_getIdentity();
    var rhsIdentity = rhs.ice_getIdentity();
    var n = lhsIdentity.name.localeCompare(rhsIdentity.name);

    if (n !== 0) {
      return n;
    }

    n = lhsIdentity.category.localeCompare(rhsIdentity.category);

    if (n !== 0) {
      return n;
    }

    var lhsFacet = lhs.ice_getFacet();
    var rhsFacet = rhs.ice_getFacet();

    if (lhsFacet === null && rhsFacet === null) {
      return 0;
    } else if (lhsFacet === null) {
      return -1;
    } else if (rhsFacet === null) {
      return 1;
    }

    return lhsFacet.localeCompare(rhsFacet);
  }
};

module.exports.Ice = Ice;